# SPDX-FileCopyrightText: 2025 RTE (https://www.rte-france.com)
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0

from typing import Any

import numpy as np

from thermohl import floatArrayLike
from thermohl.power import PowerTerm
from thermohl.power.convective_cooling import compute_wind_attack_angle
from thermohl.power.cigre import Air


class ConvectiveCooling(PowerTerm):
    """Convective cooling term."""

    def __init__(
        self,
        altitude: floatArrayLike,
        cable_azimuth: floatArrayLike,
        ambient_temperature: floatArrayLike,
        wind_speed: floatArrayLike,
        outer_diameter: floatArrayLike,
        roughness_ratio: floatArrayLike,
        wind_azimuth: floatArrayLike = None,
        wind_attack_angle: floatArrayLike = None,
        g: float = 9.81,
        **kwargs: Any,
    ):
        r"""Init with args.

        If more than one input are numpy arrays, they should have the same size.

        Args:
            altitude (float | numpy.ndarray): Altitude (m).
            cable_azimuth (float | numpy.ndarray): Azimuth (deg).
            ambient_temperature (float | numpy.ndarray): Ambient temperature (°C).
            wind_speed (float | numpy.ndarray): Wind speed (m·s⁻¹).
            wind_azimuth (float | numpy.ndarray): wind azimuth regarding north (deg).
            outer_diameter (float | numpy.ndarray): External diameter (m).
            roughness_ratio (float | numpy.ndarray): Cable roughness (—).
            g (float, optional): Gravitational acceleration (m·s⁻²). The default is 9.81.

        """
        self.altitude = altitude
        self.ambient_temp = ambient_temperature
        self.wind_speed = wind_speed
        self.outer_diameter = outer_diameter
        self.roughness_ratio = roughness_ratio
        self.gravity = g

        if wind_attack_angle is None and wind_azimuth is None:
            raise ValueError("Must provide either wind_attack_angle or wind_azimuth.")
        if wind_attack_angle is not None and wind_azimuth is not None:
            print(
                "Warning: both wind_attack_angle and wind_azimuth are provided. wind_azimuth will be ignored."
            )
        if wind_attack_angle is not None:
            self.wind_attack_angle = wind_attack_angle
        else:
            self.wind_attack_angle = compute_wind_attack_angle(
                cable_azimuth, wind_azimuth
            )

    def _nu_forced(
        self, film_temperature: floatArrayLike, kinematic_viscosity: floatArrayLike
    ) -> floatArrayLike:
        """
        Calculate the Nusselt number for forced convection.

        Args:
            film_temperature (float | numpy.ndarray): Film temperature (°C).
            kinematic_viscosity (float | numpy.ndarray): Kinematic viscosity (m²·s⁻¹).

        Returns:
            float | numpy.ndarray: Nusselt number for forced convection.

        Notes:
            The function calculates the Nusselt number based on the relative density of air,
            the Reynolds number, and empirical correlations. The correlations are adjusted
            depending on the Reynolds number and the roughness ratio roughness_ratio. The function also
            considers the angle of attack (da) to adjust the coefficients.
        """
        relative_density = Air.relative_density(film_temperature, self.altitude)
        reynolds = (
            relative_density
            * np.abs(self.wind_speed)
            * self.outer_diameter
            / kinematic_viscosity
        )

        s = (
            np.ones_like(film_temperature)
            * np.ones_like(kinematic_viscosity)
            * np.ones_like(reynolds)
        )
        z = s.shape == ()
        if z:
            s = np.array([1.0])

        B1 = 0.641 * s
        n = 0.471 * s

        # NB : (0.641/0.178)**(1/(0.633-0.471)) = 2721.4642715250125
        ix = np.logical_and(
            self.roughness_ratio <= 0.05, reynolds >= 2721.4642715250125
        )
        # NB : (0.641/0.048)**(1/(0.800-0.471)) = 2638.3210085195865
        jx = np.logical_and(self.roughness_ratio > 0.05, reynolds >= 2638.3210085195865)

        B1[ix] = 0.178
        B1[jx] = 0.048

        n[ix] = 0.633
        n[jx] = 0.800

        if z:
            B1 = B1[0]
            n = n[0]

        B2 = np.where(self.wind_attack_angle < np.deg2rad(24.0), 0.68, 0.58)
        m1 = np.where(self.wind_attack_angle < np.deg2rad(24.0), 1.08, 0.90)

        return np.maximum(0.42 + B2 * np.sin(self.wind_attack_angle) ** m1, 0.55) * (
            B1 * reynolds**n
        )

    def _nu_natural(
        self,
        film_temperature: floatArrayLike,
        temperature_delta: floatArrayLike,
        kinematic_viscosity: floatArrayLike,
    ) -> floatArrayLike:
        """
        Calculate the Nusselt number for natural convection.

        Args:
            film_temperature (float | numpy.ndarray): Film temperature (°C).
            temperature_delta (float | numpy.ndarray): Temperature difference (°C).
            kinematic_viscosity (float | numpy.ndarray): Kinematic viscosity (m²·s⁻¹).

        Returns:
            float | numpy.ndarray: Nusselt number for natural convection.

        Notes:
            The function calculates the Grashof number (gr) and the product of the Grashof
            number and the Prandtl number (gp). It then uses these values to determine the
            Nusselt number based on empirical correlations for different ranges of gp.

        """
        grashof = (
            self.outer_diameter**3
            * np.abs(temperature_delta)
            * self.gravity
            / ((film_temperature + 273.15) * kinematic_viscosity**2)
        )
        gr_prandtl = grashof * Air.prandtl(film_temperature)
        ia = gr_prandtl < 1.0e04
        A2 = np.ones_like(gr_prandtl) * 0.480
        m2 = np.ones_like(gr_prandtl) * 0.250

        if len(gr_prandtl.shape) == 0:
            if ia:
                A2 = 0.850
                m2 = 0.188
        else:
            A2[ia] = 0.850
            m2[ia] = 0.188
        return A2 * gr_prandtl**m2

    def value(self, conductor_temperature: floatArrayLike) -> floatArrayLike:
        r"""Compute convective cooling.

        Args:
            conductor_temperature (float | numpy.ndarray): Conductor temperature (°C).

        Returns:
            float | numpy.ndarray: Power term value (W·m⁻¹).

        """
        film_temperature = 0.5 * (conductor_temperature + self.ambient_temp)
        temperature_delta = conductor_temperature - self.ambient_temp
        kinematic_viscosity = Air.kinematic_viscosity(film_temperature)
        # nu[nu < 1.0E-06] = 1.0E-06
        thermal_conductivity = Air.thermal_conductivity(film_temperature)
        # lm[lm < 0.01] = 0.01
        nusselt_forced = self._nu_forced(film_temperature, kinematic_viscosity)
        nusselt_natural = self._nu_natural(
            film_temperature, temperature_delta, kinematic_viscosity
        )
        return (
            np.pi
            * thermal_conductivity
            * (conductor_temperature - self.ambient_temp)
            * np.maximum(nusselt_forced, nusselt_natural)
        )



# SPDX-FileCopyrightText: 2025 RTE (https://www.rte-france.com)
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0

from abc import ABC
from typing import Any

import numpy as np

from thermohl import floatArrayLike

_TEMP_INCREMENT = 1.0e-03


class PowerTerm(ABC):
    """Base class for power term."""

    def __init__(self, **kwargs: Any):
        pass

    def value(self, conductor_temperature: floatArrayLike) -> floatArrayLike:
        r"""Compute power term value in function of temperature.

        Usually this function should be overridden in children classes; if it is
        not the case it will just return zero.

        Args:
            conductor_temperature (float | numpy.ndarray): Conductor temperature (°C).

        Returns:
            float | numpy.ndarray: Power term value (W·m⁻¹).

        """
        return (
            np.zeros_like(conductor_temperature)
            if not np.isscalar(conductor_temperature)
            else 0.0
        )

    def derivative(
        self,
        conductor_temperature: floatArrayLike,
        temperature_increment: float = _TEMP_INCREMENT,
    ) -> floatArrayLike:
        r"""Compute power term derivative regarding temperature in function of temperature.

        Usually this function should be overriden in children classes; if it is
        not the case it will evaluate the derivative from the value method with
        a second-order approximation.

        Args:
            conductor_temperature (float | numpy.ndarray): Conductor temperature (°C).
            temperature_increment (float, optional): Temperature increment. The default is 1.0E-03.

        Returns:
            float | numpy.ndarray: Power term derivative (W·m⁻¹·K⁻¹).

        """
        return (
            self.value(conductor_temperature + temperature_increment)
            - self.value(conductor_temperature - temperature_increment)
        ) / (2.0 * temperature_increment)



# SPDX-FileCopyrightText: 2025 RTE (https://www.rte-france.com)
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0

from datetime import datetime
import numpy as np
from thermohl.sun import utc2solar_hour


def test_scalar_input():
    # Test with scalar inputs
    datetime_utc = datetime(2025, 5, 18, 12, 30, 45)
    longitude = np.deg2rad(45)  # 45 degrees east

    result = utc2solar_hour(datetime_utc, longitude)
    expected_result = 15.57397066555581
    assert np.isclose(result, expected_result)


def test_array_input():
    # Test with numpy array inputs
    datetimes_utc = [
        datetime(2025, 3, 18, 12, 30, 0),
        datetime(2025, 9, 2, 15, 45, 0),
        datetime(2025, 12, 27, 18, 0, 0),
    ]
    longitudes = np.deg2rad(np.array([0, 90, 135]))

    result = utc2solar_hour(datetimes_utc, longitudes)
    expected_result = np.array([12.35394, 21.76352, 26.976276])
    np.testing.assert_array_almost_equal(result, expected_result)



# SPDX-FileCopyrightText: 2025 RTE (https://www.rte-france.com)
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0

"""Generic radiative cooling term."""

from typing import Any

import numpy as np

from thermohl import floatArrayLike
from thermohl.power.power_term import PowerTerm


class RadiativeCoolingBase(PowerTerm):
    """Generic power term for radiative cooling."""

    def _celsius2kelvin(self, temperature: floatArrayLike) -> floatArrayLike:
        return temperature + self.kelvin_offset

    def __init__(
        self,
        ambient_temperature: floatArrayLike,
        outer_diameter: floatArrayLike,
        emissivity: floatArrayLike,
        stefan_boltzmann_constant: float = 5.67e-08,
        zerok: float = 273.15,
        **kwargs: Any,
    ):
        r"""Init with args.

        Args:
            ambient_temperature (float | numpy.ndarray): Ambient temperature (°C).
            outer_diameter (float | numpy.ndarray): External diameter (m).
            emissivity (float | numpy.ndarray): Emissivity (—).
            stefan_boltzmann_constant (float, optional): Stefan–Boltzmann constant (W·m⁻²·K⁻⁴). The default is 5.67e-08.
            zerok (float, optional): Offset to convert Celsius to Kelvin (K). The default is 273.15.

        """
        self.kelvin_offset = zerok
        self.ambient_temperature = self._celsius2kelvin(ambient_temperature)
        self.outer_diameter = outer_diameter
        self.emissivity = emissivity
        self.stefan_boltzmann = stefan_boltzmann_constant

    def value(self, conductor_temperature: floatArrayLike) -> floatArrayLike:
        r"""Compute radiative cooling using the Stefan-Boltzmann law.

        Args:
            conductor_temperature (float | numpy.ndarray): Conductor temperature (°C).

        Returns:
            float | numpy.ndarray: Power term value (W·m⁻¹).

        """
        return (
            np.pi
            * self.stefan_boltzmann
            * self.emissivity
            * self.outer_diameter
            * (
                self._celsius2kelvin(conductor_temperature) ** 4
                - self.ambient_temperature**4
            )
        )

    def derivative(self, conductor_temperature: floatArrayLike) -> floatArrayLike:
        r"""Analytical derivative of value method.

        Args:
            conductor_temperature (float | numpy.ndarray): Conductor temperature (K).

        Returns:
            float | numpy.ndarray: Power term derivative (W·m⁻¹·K⁻¹).

        """
        return (
            4.0
            * np.pi
            * self.stefan_boltzmann
            * self.emissivity
            * self.outer_diameter
            * conductor_temperature**3
        )



# SPDX-FileCopyrightText: 2025 RTE (https://www.rte-france.com)
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0

from typing import Any

from thermohl.power import ieee
from thermohl import floatArrayLike, datetimeListLike


class SolarHeating(ieee.SolarHeating):
    """Solar heating term."""

    def __init__(
        self,
        latitude: floatArrayLike,
        altitude: floatArrayLike,
        cable_azimuth: floatArrayLike,
        datetime_utc: datetimeListLike,
        outer_diameter: floatArrayLike,
        solar_absorptivity: floatArrayLike,
        measured_solar_irradiance: floatArrayLike,
        **kwargs: Any,
    ):
        """Init with args.
        See ieee.SolarHeating, it is exactly the same with altitude and turbidity set to zero.
        If more than one input are numpy arrays, they should have the same size.

        :param latitude: Latitude in degrees.
        :param altitude: Altitude.
        :param cable_azimuth: Azimuth of the conductor in degrees.
        :param datetime_utc: Datetime in UTC.
        :param outer_diameter: external diameter of the conductor.
        :param solar_absorptivity: Solar absorption coefficient of the conductor.
        :param measured_solar_irradiance: Optional precomputed solar radiation term.
        """
        if "turbidity" in kwargs.keys():
            kwargs.pop("turbidity")
        super().__init__(
            latitude=latitude,
            altitude=altitude,
            cable_azimuth=cable_azimuth,
            turbidity=0.0,
            datetime_utc=datetime_utc,
            outer_diameter=outer_diameter,
            solar_absorptivity=solar_absorptivity,
            measured_solar_irradiance=measured_solar_irradiance,
            **kwargs,
        )



# SPDX-FileCopyrightText: 2025 RTE (https://www.rte-france.com)
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0

import numbers
from typing import Dict, Any, Optional

import numpy as np
import pandas as pd

from thermohl import floatArrayLike, floatArray
from thermohl.solver.solver import Solver as Solver_, get_time_changing_parameters
from thermohl.solver.parameters import DEFAULT_PARAMETERS as default
from thermohl.solver.entities import PowerType, VariableType
from thermohl.utils import bisect_v


class Solver1T(Solver_):
    def steady_temperature(
        self,
        Tmin: float = default.tmin,
        Tmax: float = default.tmax,
        tol: float = default.tol,
        maxiter: int = default.maxiter,
        return_err: bool = False,
        return_power: bool = True,
    ) -> pd.DataFrame:
        """
        Compute steady-state temperature.

        Args:
            Tmin (float, optional): Lower bound for temperature.
            Tmax (float, optional): Upper bound for temperature.
            tol (float, optional): Tolerance for temperature error.
            maxiter (int, optional): Max number of iterations.
            return_err (bool, optional): Return final error on temperature to check convergence. The default is False.
            return_power (bool, optional): Return power term values. The default is True.

        Returns:
            pandas.DataFrame: A DataFrame with temperature and other results (depending on inputs) in the columns.

        """

        # solve with bisection
        conductor_temperature, err = bisect_v(
            lambda x: -self.balance(x),
            Tmin,
            Tmax,
            (self.args.get_number_of_computations(),),
            tol,
            maxiter,
        )

        # format output
        df = pd.DataFrame(
            data=conductor_temperature, columns=[VariableType.TEMPERATURE]
        )

        if return_err:
            df[VariableType.ERROR] = err

        if return_power:
            df[PowerType.JOULE] = self.joule_heating.value(conductor_temperature)
            df[PowerType.SOLAR] = self.solar_heating.value(conductor_temperature)
            df[PowerType.CONVECTION] = self.convective_cooling.value(
                conductor_temperature
            )
            df[PowerType.RADIATION] = self.radiative_cooling.value(
                conductor_temperature
            )
            df[PowerType.RAIN] = self.precipitation_cooling.value(conductor_temperature)

        return df

    def transient_temperature(
        self,
        offset: floatArray = np.array([]),
        T0: Optional[float] = None,
        return_power: bool = False,
    ) -> Dict[str, Any]:
        """
        Compute transient-state temperature.

        Args:
            offset (numpy.ndarray): A 1D array with times (in seconds) when the temperature needs to be computed. The array must contain increasing values (undefined behaviour otherwise).
            T0 (float | None): Initial temperature. If None, the ambient temperature from the internal dict will be used. The default is None.
            return_power (bool, optional): Return power term values. The default is False.

        Returns:
            Dict[str, Any]: A dictionary with temperature and other results (depending on inputs) in the keys.
        """

        # get sizes
        n = self.args.get_number_of_computations()
        N = len(offset)
        if N < 2:
            raise ValueError("The length of the time array must be at least 2.")

        # get initial temperature
        if T0 is None:
            T0 = (
                self.args.ambient_temperature
                if isinstance(self.args.ambient_temperature, numbers.Number)
                else self.args.ambient_temperature[0]
            )
        time_changing_parameters = get_time_changing_parameters(self.args, offset, N, n)
        # inverse of m*C : shortcuts for time-loop
        imc = 1.0 / (self.args.linear_mass * self.args.heat_capacity)

        # init
        conductor_temperature = np.zeros((N, n))
        conductor_temperature[0, :] = T0

        # main time loop
        for i in range(1, N):
            for k, v in time_changing_parameters.items():
                self.args[k] = v[i, :]
            self.update()
            conductor_temperature[i, :] = (
                conductor_temperature[i - 1, :]
                + (offset[i] - offset[i - 1])
                * self.balance(conductor_temperature[i - 1, :])
                * imc
            )

        # save results
        result = {
            VariableType.TIME: offset,
            VariableType.TEMPERATURE: conductor_temperature,
        }

        # manage return dict 2: powers
        if return_power:
            for power in Solver_.powers():
                result[power] = np.zeros_like(conductor_temperature)
            for i in range(N):
                for key in time_changing_parameters.keys():
                    self.args[key] = time_changing_parameters[key][i, :]
                self.update()
                result[PowerType.JOULE][i, :] = self.joule_heating.value(
                    conductor_temperature[i, :]
                )
                result[PowerType.SOLAR][i, :] = self.solar_heating.value(
                    conductor_temperature[i, :]
                )
                result[PowerType.CONVECTION][i, :] = self.convective_cooling.value(
                    conductor_temperature[i, :]
                )
                result[PowerType.RADIATION][i, :] = self.radiative_cooling.value(
                    conductor_temperature[i, :]
                )
                result[PowerType.RAIN][i, :] = self.precipitation_cooling.value(
                    conductor_temperature[i, :]
                )

        # squeeze return values if n is 1
        if n == 1:
            keys = list(result.keys())
            keys.remove(VariableType.TIME)
            for key in keys:
                result[key] = result[key][:, 0]

        return result

    def steady_intensity(
        self,
        max_conductor_temperature: floatArrayLike = np.array([]),
        Imin: float = default.imin,
        Imax: float = default.imax,
        tol: float = default.tol,
        maxiter: int = default.maxiter,
        return_err: bool = False,
        return_power: bool = True,
    ) -> pd.DataFrame:
        """Compute steady-state max intensity.

        Compute the maximum intensity that can be run in a conductor without
        exceeding the temperature given in argument.

        Args:
            max_conductor_temperature (float | numpy.ndarray): Maximum temperature.
            Imin (float, optional): Lower bound for intensity. The default is 0.
            Imax (float, optional): Upper bound for intensity. The default is 9999.
            tol (float, optional): Tolerance for temperature error. The default is 1.0E-06.
            maxiter (int, optional): Max number of iterations. The default is 64.
            return_err (bool, optional): Return final error on intensity to check convergence. The default is False.
            return_power (bool, optional): Return power term values. The default is True.

        Returns:
            pandas.DataFrame: A dataframe with maximum intensity and other results (depending on inputs) in the columns.

        """

        # save transit in arg
        transit = self.args.transit

        # solve with bisection
        shape = (self.args.get_number_of_computations(),)
        T_ = max_conductor_temperature * np.ones(shape)
        joule_heating = (
            self.convective_cooling.value(T_)
            + self.radiative_cooling.value(T_)
            + self.precipitation_cooling.value(T_)
            - self.solar_heating.value(T_)
        )

        def fun(i: floatArray) -> floatArrayLike:
            self.args.transit = i
            self.joule_heating.__init__(**self.args.__dict__)
            return self.joule_heating.value(T_) - joule_heating

        A, err = bisect_v(fun, Imin, Imax, shape, tol, maxiter)

        # restore previous transit
        self.args.transit = transit

        # format output
        df = pd.DataFrame(data=A, columns=[VariableType.TRANSIT])

        if return_err:
            df[VariableType.ERROR] = err

        if return_power:
            df[PowerType.JOULE] = self.joule_heating.value(max_conductor_temperature)
            df[PowerType.SOLAR] = self.solar_heating.value(max_conductor_temperature)
            df[PowerType.CONVECTION] = self.convective_cooling.value(
                max_conductor_temperature
            )
            df[PowerType.RADIATION] = self.radiative_cooling.value(
                max_conductor_temperature
            )
            df[PowerType.RAIN] = self.precipitation_cooling.value(
                max_conductor_temperature
            )

        return df



# SPDX-FileCopyrightText: 2025 RTE (https://www.rte-france.com)
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0

"""Functions and classes to obtain various probability distributions for scalar random variables."""

import warnings
from typing import Union, Optional

import numpy as np

from thermohl.utils import depends_on_optional

try:
    import scipy
    from scipy.special import erf, i0, i1
    from scipy.stats._distn_infrastructure import rv_continuous_frozen as frozen_dist
except ImportError:
    warnings.warn(
        "scipy is not installed. Some functions will not be available.",
        RuntimeWarning,
    )
    frozen_dist = object


from thermohl import floatArrayLike

_twopi = 2 * np.pi


def _phi(value: float) -> float:
    """PDF of standard normal distribution."""
    return np.exp(-0.5 * value**2) / np.sqrt(_twopi)


@depends_on_optional("scipy")
def _psi(value: float) -> float:
    """CDF of standard normal distribution."""
    return 0.5 * (1 + erf(value / np.sqrt(2)))


def _truncnorm_header(
    lower_bound: float, upper_bound: float, mean: float, standard_deviation: float
) -> tuple[float, float, float, float]:
    """Utility code factoring."""
    alpha = (lower_bound - mean) / standard_deviation
    beta = (upper_bound - mean) / standard_deviation
    return alpha, beta, mean, standard_deviation


def _truncnorm_mean_std(
    lower_bound: float, upper_bound: float, mean: float, standard_deviation: float
) -> tuple[float, float]:
    """Real mean and std of truncated normal distribution."""
    alpha, beta, mean, standard_deviation = _truncnorm_header(
        lower_bound, upper_bound, mean, standard_deviation
    )
    normalizer = _psi(beta) - _psi(alpha)
    mean_value = mean + standard_deviation * (_phi(alpha) - _phi(beta)) / normalizer
    standard_value = standard_deviation * np.sqrt(
        1
        + (alpha * _phi(alpha) - beta * _phi(beta)) / normalizer
        - ((_phi(alpha) - _phi(beta)) / normalizer) ** 2
    )
    return mean_value, standard_value


@depends_on_optional("scipy")
def truncnorm(
    lower_bound: float,
    upper_bound: float,
    mean: float,
    standard_deviation: float,
    mean_tolerance: float = 1.0e-03,
    std_tolerance: float = 1.0e-02,
    relative_tolerance: bool = True,
) -> frozen_dist:
    """Truncated normal distribution. Wrapper from scipy.stats."""
    if lower_bound >= upper_bound:
        raise ValueError(
            "Input lower_bound (%.3E) should be lower than upper_bound (%.3E)."
            % (lower_bound, upper_bound)
        )
    if mean < lower_bound or mean > upper_bound:
        raise ValueError(
            "Input mean (%.3E) should be in [lower_bound, upper_bound] range (%.3E, %.3E)."
            % (mean, lower_bound, upper_bound)
        )
    if standard_deviation < 0.0:
        raise ValueError(
            "Input standard_deviation (%.3E) should be positive."
            % (standard_deviation,)
        )

    target_mean = mean
    target_standard_deviation = standard_deviation
    alpha, beta, mean, standard_deviation = _truncnorm_header(
        lower_bound, upper_bound, target_mean, target_standard_deviation
    )
    distribution = scipy.stats.truncnorm(alpha, beta, mean, standard_deviation)

    actual_mean = distribution.mean()
    actual_standard_deviation = distribution.std()

    mean_tol = mean_tolerance
    std_tol = std_tolerance
    if relative_tolerance:
        mean_tol *= target_mean
        std_tol *= target_standard_deviation
    if np.abs(target_standard_deviation - actual_standard_deviation) >= std_tol:
        warnings.warn(
            "Required std cannot be achieved (%.3E instead of %.3E). Choose a lower std, extend your "
            "bounds or change your distribution."
            % (target_standard_deviation, actual_standard_deviation),
            RuntimeWarning,
        )
    if np.abs(target_mean - actual_mean) >= mean_tol:
        warnings.warn(
            "Required mean cannot be achieved (%.3E instead of %.3E). Move the mean towards the center of your "
            "bounds, extend your bounds or change your distribution."
            % (target_mean, actual_mean),
            RuntimeWarning,
        )

    return distribution


@depends_on_optional("scipy")
class WrappedNormal(object):
    """Wrapped-Normal distribution. Not as complete as a scipy.stat distribution."""

    def __init__(
        self, mean: float, standard_deviation: float, lower_bound: float = 0.0
    ):
        if standard_deviation < 0:
            raise ValueError("Std should be positive.")
        if mean < lower_bound or mean >= lower_bound + _twopi:
            raise ValueError(
                "Mean should be greater than lower bound and lower than lower bound + 2*pi."
            )
        self.lower_bound = lower_bound
        self.upper_bound = lower_bound + _twopi
        self.mean_value = mean
        self.standard_deviation = standard_deviation
        return

    @depends_on_optional("scipy")
    def rvs(
        self,
        size: int = 1,
        random_state: Optional[
            Union[int, np.random.Generator, np.random.RandomState]
        ] = None,
    ) -> floatArrayLike:
        sample = scipy.stats.norm.rvs(
            loc=self.mean_value,
            scale=self.standard_deviation,
            size=size,
            random_state=random_state,
        )
        sample = sample % _twopi
        sample[sample < self.lower_bound] += _twopi
        sample[sample > self.upper_bound] -= _twopi
        return sample

    def mean(self) -> float:
        return self.mean_value

    def median(self) -> float:
        return self.mean_value

    def var(self) -> float:
        return 1 - np.exp(-0.5 * self.standard_deviation**2)

    def std(self) -> float:
        return np.sqrt(self.var())

    def ppf(self, q: float) -> np.float64:
        return np.quantile(self.rvs(9999), q)


def wrapnorm(mean: float, standard_deviation: float) -> WrappedNormal:
    """Get Wrapped Normal distribution.
    -- in radians, in [0, 2*pi]
    """
    mean_wrapped = mean % _twopi
    if mean_wrapped != mean:
        warnings.warn(
            "Changed mean from %.3E to %.3E to fit [0,2*pi] interval."
            % (mean, mean_wrapped),
            RuntimeWarning,
        )
    if standard_deviation >= _twopi:
        warnings.warn(
            "Required std cannot be achieved (%.3E > 2*pi). Choose a lower std "
            "or change your distribution." % (standard_deviation,),
            RuntimeWarning,
        )
    return WrappedNormal(mean_wrapped, standard_deviation)


@depends_on_optional("scipy")
def _vonmises_circ_var(kappa: float) -> float:
    """Von Mises distribution circular variance."""
    return 1 - i1(kappa) / i0(kappa)


@depends_on_optional("scipy")
def _vonmises_kappa(standard_deviation: float) -> float:
    """Get von Mises parameter that matches std in input."""
    from scipy.optimize import newton

    circ_variance = 1.0 - np.exp(-0.5 * standard_deviation**2)
    kappa_guess = 0.5 / circ_variance

    def fun(x: float) -> float:
        return _vonmises_circ_var(x) - circ_variance

    try:
        kappa = newton(fun, x0=kappa_guess, tol=1.0e-06, maxiter=32)
    except RuntimeError:
        kappa = 1 / standard_deviation**2

    return kappa


@depends_on_optional("scipy")
def vonmises(mean: float, standard_deviation: float) -> frozen_dist:
    """Get von Mises distribution.
    -- in radians, in [-pi,+pi]
    """
    mean_wrapped = mean % _twopi
    if mean_wrapped > np.pi:
        mean_wrapped -= _twopi
    if mean_wrapped != mean:
        warnings.warn(
            "Changed mean from %.3E to %.3E to fit [-pi,+pi] interval."
            % (mean, mean_wrapped),
            RuntimeWarning,
        )
    if standard_deviation < 0.0:
        raise ValueError(
            "Input standard_deviation (%.3E) should be positive."
            % (standard_deviation,)
        )
    sigmax = _twopi / np.sqrt(12)
    if standard_deviation >= sigmax:
        warnings.warn(
            "Required std cannot be achieved (%.3E > %.3E). Choose a lower std "
            "or change your distribution." % (standard_deviation, sigmax),
            RuntimeWarning,
        )
    kappa = _vonmises_kappa(standard_deviation)
    return scipy.stats.vonmises_line(kappa, loc=mean_wrapped)
